#!/usr/bin/perl -w

# Check for libraries required by ncad that aren't present on the root
# filesystem and copy them into /lib.  If any argument at all is
# given, just display the copy operations that would have been made,
# without actually making them.

# Julian T. J. Midgley <jtjm@xenoclast.org>

use strict;

my $DISP_ONLY = scalar(@ARGV);
my %required_libs;

open (LIBS, "ldd /sbin/ncad |") 
    or die "ldd /sbin/ncad failed: $!\n";
while (<LIBS>) {
    if ( /(\S+)\s+=>\s+(\S+)/ ) {
        $required_libs{$2} = $1;
    }
    next;
}

my @non_root_fs;

open (FSTAB, "/etc/fstab") 
    or die "Unable to read /etc/fstab: $!\n";
while (<FSTAB>) {
    next if (/^\#/);
    if ( /^(\S+)\s+(\/\S+)/ ) {
        push ( @non_root_fs, $2 );
    }
}

if ($DISP_ONLY) {
    print "Display mode: not executing commands\n";
}

foreach my $lib (keys %required_libs) {
    my $part = (split(/\//, $lib))[1];
    if (grep (/^\/$part$/, @non_root_fs)) {
        my $cmd = "/bin/cp";
        my @args = ($lib, "/lib/$required_libs{$lib}");
        print "  $cmd " . join(" ",@args) . "\n";
        unless ($DISP_ONLY) {
            system ($cmd, @args);
        }
    }
}
    

